from typing import Any, Dict, List, Optional

import rzpipe

from fwhunt_scan.uefi_protocols import UefiGuid as UefiGuid

def get_int(item: str) -> Optional[int]: ...
def get_xrefs_to_guids(rz: rzpipe.open, guids: List[UefiGuid]) -> List[int]: ...
def get_xrefs_to_data(rz: rzpipe.open, addr: int) -> List[int]: ...
def get_current_insn_index(insns: Optional[List[Dict[str, Any]]], code_addr: int) -> Optional[int]: ...
