from typing import Dict

from _typeshed import Incomplete

from fwhunt_scan.uefi_types import UefiGuid as UefiGuid

PROTOCOLS_GUIDS: Incomplete
GUID_FROM_VALUE: Dict[str, UefiGuid]
GUID_FROM_BYTES: Dict[bytes, UefiGuid]
